(*
Author:  Akihisa Yamada (2018-2019)
License: LGPL (see file COPYING.LESSER)
*)
theory Binary_Relations
  imports
Main
    HOL.Complete_Partial_Order HOL.Wellfounded HOL.Conditionally_Complete_Lattices
begin

no_notation conversep ("(_\<inverse>\<inverse>)" [1000] 1000)

section {* Binary Relations *}

(* TODO: why is this hidden? *)
abbreviation "chain \<equiv> Complete_Partial_Order.chain"

no_notation conversep ("(_\<inverse>\<inverse>)" [1000] 1000)

abbreviation(input) dual ("(_\<inverse>\<inverse>)" [1000] 1000) where "dual rel x y \<equiv> rel y x"

lemma conversep_as_dual[simp]: "conversep r = r\<inverse>\<inverse>" by auto

lemma fun_ordI: "(\<And>x. r (f x) (g x)) \<Longrightarrow> fun_ord r f g"
  and fun_ordD: "fun_ord r f g \<Longrightarrow> r (f x) (g x)"
  by (auto simp: fun_ord_def)

syntax
  "_range" :: "'a \<Rightarrow> pttrn \<Rightarrow> 'a set" ("(1{_ /|./ _})")
  "_image" :: "'a \<Rightarrow> pttrn \<Rightarrow> 'a set \<Rightarrow> 'a set"  ("(1{_ /|./ (_/ \<in> _)})")
translations
  "{e |. p}" \<rightleftharpoons> "CONST range (\<lambda>p. e)"
  "{e |. p \<in> A}" \<rightleftharpoons> "CONST image (\<lambda>p. e) A"

lemma image_constant: "(\<And>i. i \<in> I \<Longrightarrow> f i = y) \<Longrightarrow> f`I = (if I = {} then {} else {y})"
  by auto


subsubsection {* Various Definitions *}

lemma monotone_dual: "monotone r s f \<Longrightarrow> monotone r\<inverse>\<inverse> s\<inverse>\<inverse> f"
  by (auto simp: monotone_def)

lemma monotone_id: "monotone r r id"
  by (auto simp: monotone_def)

context fixes r :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<sqsubseteq>" 50) begin

definition "bound X b \<equiv> \<forall>x \<in> X. x \<sqsubseteq> b"

lemma boundI[intro!]: "(\<And>x. x \<in> X \<Longrightarrow> x \<sqsubseteq> b) \<Longrightarrow> bound X b"
  and boundE[elim]: "bound X b \<Longrightarrow> ((\<And>x. x \<in> X \<Longrightarrow> x \<sqsubseteq> b) \<Longrightarrow> thesis) \<Longrightarrow> thesis"
  by (auto simp: bound_def)

lemma bound_empty[simp]: "bound {} = (\<lambda>x. True)" by auto
lemma bound_insert[simp]: "bound (insert x X) b \<longleftrightarrow> x \<sqsubseteq> b \<and> bound X b" by auto

lemma bound_cmono: assumes "X \<subseteq> Y" shows "bound Y x \<Longrightarrow> bound X x"
  using assms by auto

definition reflexive_on where "reflexive_on X \<equiv> \<forall>x \<in> X. x \<sqsubseteq> x"

lemma reflexive_onI[intro]:
  assumes "\<And>x. x \<in> X \<Longrightarrow> x \<sqsubseteq> x" shows "reflexive_on X" using assms reflexive_on_def by auto

lemma reflexive_onE[elim]:
  assumes "reflexive_on X" and "(\<And>x. x \<in> X \<Longrightarrow> x \<sqsubseteq> x) \<Longrightarrow> thesis" shows thesis
  using assms reflexive_on_def by auto

lemma chain_imp_reflexive: "chain (\<sqsubseteq>) X \<Longrightarrow> reflexive_on X" by (auto elim: chainE)

definition "directed X \<equiv> \<forall>x \<in> X. \<forall> y \<in> X. \<exists>z \<in> X. x \<sqsubseteq> z \<and> y \<sqsubseteq> z"

lemma directedE:
  assumes "directed X" and "x \<in> X" and "y \<in> X"
    and "\<And>z. z \<in> X \<Longrightarrow> x \<sqsubseteq> z \<Longrightarrow> y \<sqsubseteq> z \<Longrightarrow> thesis"
  shows "thesis"
  using assms by (auto simp: directed_def)

lemma directedI[intro]:
  assumes "\<And>x y. x \<in> X \<Longrightarrow> y \<in> X \<Longrightarrow> \<exists>z \<in> X. x \<sqsubseteq> z \<and> y \<sqsubseteq> z"
  shows "directed X" 
  using assms by (auto simp: directed_def)

lemma chain_imp_directed: "chain (\<sqsubseteq>) X \<Longrightarrow> directed X"
  by (intro directedI, auto elim: chainE)

definition "extreme X e \<equiv> e \<in> X \<and> (\<forall>x \<in> X. x \<sqsubseteq> e)"

lemma extremeI[intro]: "e \<in> X \<Longrightarrow> (\<And>x. x \<in> X \<Longrightarrow> x \<sqsubseteq> e) \<Longrightarrow> extreme X e"
  and extremeD: "extreme X e \<Longrightarrow> e \<in> X" "extreme X e \<Longrightarrow> (\<And>x. x \<in> X \<Longrightarrow> x \<sqsubseteq> e)"
  and extremeE[elim]: "extreme X e \<Longrightarrow> (e \<in> X \<Longrightarrow> (\<And>x. x \<in> X \<Longrightarrow> x \<sqsubseteq> e) \<Longrightarrow> thesis) \<Longrightarrow> thesis"
  by (auto simp: extreme_def)

lemma extreme_UNIV[simp]: "extreme UNIV t \<longleftrightarrow> (\<forall>x. x \<sqsubseteq> t)" by auto

lemma extremes_equiv: "extreme X b \<Longrightarrow> extreme X c \<Longrightarrow> b \<sqsubseteq> c \<and> c \<sqsubseteq> b" by auto

end

context fixes r :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<sqsubseteq>" 50)
begin

abbreviation "extreme_bound X \<equiv> extreme (\<lambda>x y. y \<sqsubseteq> x) {b. bound (\<sqsubseteq>) X b}"

lemma extreme_boundI[intro]:
  assumes "\<And>b. bound (\<sqsubseteq>) X b \<Longrightarrow> s \<sqsubseteq> b" and "\<And>x. x \<in> X \<Longrightarrow> x \<sqsubseteq> s"
  shows "extreme_bound X s"
  using assms by auto

lemma extreme_bound_mono:
  assumes XY: "X \<subseteq> Y"
    and bX: "extreme_bound X bX"
    and bY: "extreme_bound Y bY"
  shows "bX \<sqsubseteq> bY"
proof-
  have "bound (\<sqsubseteq>) X bY" using XY bY by force
  with bX show ?thesis by auto
qed

lemma extreme_bound_iff:
  shows "extreme_bound X b \<longleftrightarrow> (\<forall>c. (\<forall>x\<in>X. x \<sqsubseteq> c) \<longrightarrow> b \<sqsubseteq> c) \<and> (\<forall>x \<in> X. x \<sqsubseteq> b)"
  by (auto simp: extreme_def)

lemma extreme_bound_singleton[simp]:
  "extreme_bound {x} x \<longleftrightarrow> x \<sqsubseteq> x" by auto

lemma extreme_bound_equiv: "extreme_bound X b \<Longrightarrow> c \<in> X \<Longrightarrow> b \<sqsubseteq> c \<Longrightarrow> c \<sqsubseteq> b"
  by auto

lemma extreme_bound_image_const:
  "x \<sqsubseteq> x \<Longrightarrow> C \<noteq> {} \<Longrightarrow> (\<And>i. i \<in> C \<Longrightarrow> f i = x) \<Longrightarrow> extreme_bound (f ` C) x"
  by (auto simp: image_constant)

lemma extreme_bound_UN_const:
  "x \<sqsubseteq> x \<Longrightarrow> C \<noteq> {} \<Longrightarrow> (\<And>i y. i \<in> C \<Longrightarrow> P i y \<longleftrightarrow> x = y) \<Longrightarrow>
  extreme_bound (\<Union>i\<in>C. {y. P i y}) x"
  by auto

end

context
  fixes ir (infix "\<preceq>" 50)
  fixes r :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<sqsubseteq>" 50)
  fixes f e C
  assumes mono: "monotone (\<preceq>) (\<sqsubseteq>) f" and e: "extreme (\<preceq>) C e"
begin

lemma monotone_extreme_imp_extreme_bound:
  shows "extreme_bound (\<sqsubseteq>) (f ` C) (f e)"
  using monotoneD[OF mono] e
  by (auto simp: image_def intro!:extreme_boundI elim!:extremeE boundE)

lemma monotone_extreme_extreme_boundI:
  "x = f e \<Longrightarrow> extreme_bound (\<sqsubseteq>) (f ` C) x"
  using monotone_extreme_imp_extreme_bound by auto

end

lemma fun_extreme_bound_iff:
  fixes r :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<sqsubseteq>" 50)
  shows "extreme_bound (fun_ord (\<sqsubseteq>)) F e \<longleftrightarrow> (\<forall>x. extreme_bound (\<sqsubseteq>) {f x |. f \<in> F} (e x))" (is "?l \<longleftrightarrow> ?r")
proof(intro iffI allI extreme_boundI fun_ordI)
  fix f x
  assume ?r
  then have e: "extreme_bound (\<sqsubseteq>) {f x |. f \<in> F} (e x)" by auto
  show "f \<in> F \<Longrightarrow> f x \<sqsubseteq> e x" using extremeD(1)[OF e] by auto
  assume "bound (fun_ord (\<sqsubseteq>)) F f"
  then have "bound (\<sqsubseteq>) {f x |. f \<in> F} (f x)" by (auto simp: fun_ord_def)
  with e show "e x \<sqsubseteq> f x" by auto
next
  fix x y
  assume l: ?l
  from l have e: "f \<in> F \<Longrightarrow> f x \<sqsubseteq> e x" for f by (auto dest!:extremeD simp: fun_ord_def)
  then show "y \<in> {f x |. f \<in> F} \<Longrightarrow> y \<sqsubseteq> e x" by auto
  assume "bound (\<sqsubseteq>) {f x |. f \<in> F} y"
  with extremeD(1)[OF l] have "bound (fun_ord (\<sqsubseteq>)) F (e(x:=y))" by (auto simp: fun_ord_def elim!:boundE)
  with l have "fun_ord (\<sqsubseteq>) e (e(x:=y))" by auto
  from fun_ordD[OF this, of x]
  show "e x \<sqsubseteq> y" by auto
qed

context
  fixes r :: "'i \<Rightarrow> 'i \<Rightarrow> bool" (infix "\<preceq>" 50)
  fixes r' :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<sqsubseteq>" 50)
begin

lemma monotone_chain_image:
  assumes mono: "monotone (\<preceq>) (\<sqsubseteq>) f" and chain: "chain (\<preceq>) C"
  shows "chain (\<sqsubseteq>) (f ` C)"
proof (rule chainI)
  fix x y
  assume "x \<in> f ` C" and "y \<in> f ` C"
  then obtain i j where ij: "i \<in> C" "j \<in> C" and [simp]: "x = f i" "y = f j" by auto
  from chain ij have "i \<preceq> j \<or> j \<preceq> i" by (auto elim: chainE)
  with ij mono show "x \<sqsubseteq> y \<or> y \<sqsubseteq> x" by (elim disjE, auto dest: monotoneD) 
qed

lemma monotone_directed_image:
  assumes mono: "monotone (\<preceq>) (\<sqsubseteq>) f" and dir: "directed (\<preceq>) D"
  shows "directed (\<sqsubseteq>) (f ` D)"
proof (rule directedI, safe)
  fix x y assume "x \<in> D" and "y \<in> D"
  with dir obtain z where z: "z \<in> D" and "x \<preceq> z" and "y \<preceq> z" by (auto elim: directedE)
  with mono have "f x \<sqsubseteq> f z" and "f y \<sqsubseteq> f z" by (auto dest: monotoneD)
  with z show "\<exists>fz \<in> f ` D. f x \<sqsubseteq> fz \<and> f y \<sqsubseteq> fz" by auto
qed

end

subsection {* Locales for Binary Relations *}

subsubsection {* Syntactic Locales *}

locale less_eq_syntax = fixes less_eq :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<sqsubseteq>" 50)

locale less_syntax = fixes less :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<sqsubset>" 50)

locale equivalence_syntax = fixes equiv :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<sim>" 50)
begin

abbreviation equiv_class ("[_]\<^sub>\<sim>") where "[x]\<^sub>\<sim> \<equiv> { y. x \<sim> y }"

end

text {* To avoid needless constants, do not declare dual of this locale.
 *}
locale less_eq_dualize = less_eq_syntax
begin

abbreviation (input) greater_eq (infix "\<sqsupseteq>" 50) where "x \<sqsupseteq> y \<equiv> y \<sqsubseteq> x"
abbreviation (input) equiv (infix "\<sim>" 50) where "x \<sim> y \<equiv> x \<sqsubseteq> y \<and> y \<sqsubseteq> x"

lemma equiv_sym[sym]: "x \<sim> y \<Longrightarrow> y \<sim> x" by auto

end

locale less_dualize = less_syntax
begin

abbreviation (input) greater (infix "\<sqsupset>" 50) where "x \<sqsupset> y \<equiv> y \<sqsubset> x"

end

subsubsection {* Basic Properties of Relations *}

locale reflexive = less_eq_syntax + assumes refl[iff]: "x \<sqsubseteq> x"
begin

lemma eq_imp: "x = y \<Longrightarrow> x \<sqsubseteq> y" by auto

lemma extreme_singleton[simp]: "extreme (\<sqsubseteq>) {a} b \<longleftrightarrow> a = b" by auto

lemma extreme_bound_singleton[iff]: "extreme_bound (\<sqsubseteq>) {a} a" by auto

end
lemmas [intro] = reflexive.intro

locale irreflexive = less_syntax + assumes irrefl[iff]: "\<not> x \<sqsubset> x"
begin
lemma implies_not_eq: "a \<sqsubset> b \<Longrightarrow> a \<noteq> b" by auto
end
lemmas [intro] = irreflexive.intro

locale transitive = less_eq_syntax + assumes trans[trans]: "x \<sqsubseteq> y \<Longrightarrow> y \<sqsubseteq> z \<Longrightarrow> x \<sqsubseteq> z"
lemmas [intro?] = transitive.intro

locale symmetric = equivalence_syntax + assumes sym[sym]: "x \<sim> y \<Longrightarrow> y \<sim> x"
begin
lemma dual_sym: "(\<sim>)\<inverse>\<inverse> = (\<sim>)" using sym by auto
end
lemmas [intro] = symmetric.intro

locale antisymmetric = less_eq_syntax + assumes antisym[dest]: "a \<sqsubseteq> b \<Longrightarrow> b \<sqsubseteq> a \<Longrightarrow> a = b"
begin

interpretation less_eq_dualize. 

lemma equiv_iff_eq_refl: "a \<sim> b \<longleftrightarrow> a = b \<and> b \<sqsubseteq> b" by auto

lemma extreme_unique: "extreme (\<sqsubseteq>) X x \<Longrightarrow> extreme (\<sqsubseteq>) X y \<longleftrightarrow> x = y"
  by (auto elim!: extremeE)

lemma ex_extreme_iff_ex1: "Ex (extreme (\<sqsubseteq>) X) \<longleftrightarrow> Ex1 (extreme (\<sqsubseteq>) X)" by (auto simp: extreme_unique)

lemma ex_extreme_iff_the:
   "Ex (extreme (\<sqsubseteq>) X) \<longleftrightarrow> extreme (\<sqsubseteq>) X (The (extreme (\<sqsubseteq>) X))"
  apply (rule iffI)
  apply (rule  theI')
  using extreme_unique by auto

end
lemmas [intro] = antisymmetric.intro

locale asymmetric = irreflexive + strict: antisymmetric "(\<sqsubset>)"
begin
lemma asym: "a \<sqsubset> b \<Longrightarrow> b \<sqsubset> a \<Longrightarrow> False" by auto
end

lemma asymmetricI[intro]:
  fixes r :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<sqsubset>" 50)
  assumes "\<And>a b. a \<sqsubset> b \<Longrightarrow> b \<sqsubset> a \<Longrightarrow> False" 
  shows "asymmetric (\<sqsubset>)"
  apply unfold_locales using assms by auto

locale total = less_eq_syntax + assumes total: "x \<sqsubseteq> y \<or> y \<sqsubseteq> x"
begin

sublocale reflexive using total by (unfold_locales, auto)

lemma total_cases[case_names le ge]: "(x \<sqsubseteq> y \<Longrightarrow> P) \<Longrightarrow> (y \<sqsubseteq> x \<Longrightarrow> P) \<Longrightarrow> P" using total by auto

lemma chain_UNIV: "chain (\<sqsubseteq>) UNIV" by (intro chainI total)

end
lemmas [intro] = total.intro

locale linear = less_syntax + assumes linear: "x \<sqsubset> y \<or> x = y \<or> y \<sqsubset> x"
begin

lemma linear_cases[case_names less eq greater]:
  "(x \<sqsubset> y \<Longrightarrow> P) \<Longrightarrow> (x = y \<Longrightarrow> P) \<Longrightarrow> (y \<sqsubset> x \<Longrightarrow> P) \<Longrightarrow> P"
  using linear by auto

end
lemmas [intro] = linear.intro

locale well_founded = less_syntax +
  assumes induct: "\<And>P a. (\<And>x. (\<And>y. y \<sqsubset> x \<Longrightarrow> P y) \<Longrightarrow> P x) \<Longrightarrow> P a"
begin

lemma wfP[intro!]: "wfP (\<sqsubset>)" using induct wfPUNIVI by blast

sublocale asymmetric
proof (intro asymmetricI notI)
  show "a \<sqsubset> b \<Longrightarrow> b \<sqsubset> a \<Longrightarrow> False" for a b by (induct a arbitrary: b rule: induct)
qed

end

locale pseudo_order = reflexive + antisymmetric
begin

lemma equiv_eq[simp]: "a \<sqsubseteq> b \<and> b \<sqsubseteq> a \<longleftrightarrow> a = b" by auto

lemma extreme_bound_singleton_eq[simp]: "extreme_bound (\<sqsubseteq>) {x} y \<longleftrightarrow> x = y" by auto

lemma eq_iff: "x = y \<longleftrightarrow> x \<sqsubseteq> y \<and> y \<sqsubseteq> x" by auto

lemma extreme_order_iff_eq[simp]: "extreme (\<sqsubseteq>) {x. x \<sqsubseteq> e} s \<longleftrightarrow> e = s" by auto

end

locale quasi_order = reflexive + transitive

locale near_order = antisymmetric + transitive

locale partial_order = quasi_order + antisymmetric

sublocale partial_order \<subseteq> pseudo_order + near_order..

locale strict_order = irreflexive + transitive "(\<sqsubset>)"

sublocale strict_order \<subseteq> asymmetric by (auto dest: trans)
sublocale strict_order \<subseteq> near_order "(\<sqsubset>)" ..

locale well_founded_order = well_founded + transitive "(\<sqsubset>)"

sublocale well_founded_order \<subseteq> strict_order..

locale tolerance = equivalence_syntax + reflexive "(\<sim>)" + symmetric "(\<sim>)"

proposition nontolerance:
  fixes r (infix "\<bowtie>" 50)
  shows "irreflexive (\<bowtie>) \<and> symmetric (\<bowtie>) \<longleftrightarrow> tolerance (\<lambda>x y. \<not> x \<bowtie> y)"
proof safe
  assume "irreflexive (\<bowtie>)" and "symmetric (\<bowtie>)"
  then interpret irreflexive "(\<bowtie>)" + symmetric "(\<bowtie>)".
  show "tolerance (\<lambda>x y. \<not> x \<bowtie> y)" by (unfold_locales, auto dest: sym)
next
  assume "tolerance (\<lambda>x y. \<not> x \<bowtie> y)"
  then interpret tolerance "\<lambda>x y. \<not> x \<bowtie> y".
  show "irreflexive (\<bowtie>)" by auto
  show "symmetric (\<bowtie>)" using sym by auto
qed

locale partial_equivalence = equivalence_syntax + symmetric "(\<sim>)" + transitive "(\<sim>)"

locale equivalence = equivalence_syntax + symmetric "(\<sim>)" + quasi_order "(\<sim>)"

sublocale equivalence \<subseteq> partial_equivalence..

proposition reflexive_irreflexive_is_empty:
  assumes "reflexive r" and "irreflexive r"
  shows "r = (\<lambda>x y. False)"
proof(intro ext iffI)
  interpret irreflexive r + reflexive r using assms by auto
  fix x y
  assume "r x y"
  with irrefl have "x \<noteq> y" by auto
  with refl show False by auto
qed auto

proposition symmetric_antisymmetric_imp_eq:
  assumes "symmetric r" and "antisymmetric r"
  shows "r x y \<Longrightarrow> x = y"
proof-
  interpret symmetric r + antisymmetric r using assms by auto
  fix x y
  assume "r x y"
  with sym[OF this] show "x = y" by auto
qed

proposition irreflexive_transitve_symmetric_is_empty:
  assumes "irreflexive r" and "transitive r" and "symmetric r"
  shows "r = (\<lambda>x y. False)"
proof(intro ext iffI)
  interpret strict_order r using assms by (unfold strict_order_def, auto)
  interpret symmetric r using assms by auto
  fix x y
  assume "r x y"
  also note sym[OF this]
  finally have "r x x".
  then show False by auto
qed auto

locale linear_irreflexive = linear + irreflexive
begin

lemma neq_iff: "x \<noteq> y \<longleftrightarrow> x \<sqsubset> y \<or> y \<sqsubset> x" by (cases x y rule: linear_cases, auto)

end

locale total_pseudo_order = total + antisymmetric
begin

sublocale pseudo_order..

lemma not_weak_iff: "\<not> y \<sqsubseteq> x \<longleftrightarrow> x \<sqsubseteq> y \<and> x \<noteq> y" by (cases x y rule:total_cases, auto)

sublocale weak: linear "(\<sqsubseteq>)" by (auto simp: not_weak_iff)

end

locale total_quasi_order = total + transitive
begin

sublocale quasi_order..

end

text {* Total weak order and linear strict order define each other, so we will formalize them
  after giving locales for pair of weak and strict parts.
*}

locale compatible_ordering = less_eq_syntax + less_syntax + reflexive "(\<sqsubseteq>)" + irreflexive "(\<sqsubset>)" +
  constrains less :: "'a \<Rightarrow> _"
  assumes weak_strict_trans[trans]: "x \<sqsubseteq> y \<Longrightarrow> y \<sqsubset> z \<Longrightarrow> x \<sqsubset> z"
  assumes strict_weak_trans[trans]: "x \<sqsubset> y \<Longrightarrow> y \<sqsubseteq> z \<Longrightarrow> x \<sqsubset> z"
  assumes strict_implies_weak: "x \<sqsubset> y \<Longrightarrow> x \<sqsubseteq> y"
begin

text {* The strict part is necessarily transitive. *}

sublocale strict: transitive "(\<sqsubset>)"
  using weak_strict_trans[OF strict_implies_weak] by unfold_locales

interpretation strict_order..

sublocale strict: near_order "(\<sqsubset>)" by unfold_locales

sublocale asymmetric by unfold_locales

sublocale strict_order..

lemma strict_implies_not_weak: "x \<sqsubset> y \<Longrightarrow> \<not> y \<sqsubseteq> x"
  by (auto dest: strict_weak_trans)

end

locale pseudo_ordering = compatible_ordering + antisymmetric
begin

sublocale pseudo_order..

end

locale quasi_ordering = compatible_ordering + transitive
begin

sublocale quasi_order..

end

locale partial_ordering = compatible_ordering + antisymmetric + transitive
begin

sublocale partial_order + pseudo_ordering + quasi_ordering..

end

locale well_founded_ordering = quasi_ordering + well_founded
begin

end

text {* Following Isabelle/HOL's @{locale ordering}, next one fixes strict part from weak part. *}

locale ordering = partial_order + less_syntax +
  constrains less :: "'a \<Rightarrow> _"
  assumes strict_as_weak: "x \<sqsubset> y \<longleftrightarrow> x \<sqsubseteq> y \<and> x \<noteq> y"
begin

sublocale compatible_ordering by (unfold_locales, auto simp: strict_as_weak dest: trans)

sublocale partial_ordering..

end

text {* Confirming our ordering locale is equivalent to that of Isabelle/HOL. *}

sublocale Orderings.ordering \<subseteq> new: ordering
  by (unfold_locales, auto dest: trans antisym simp: strict_iff_order refl)

sublocale ordering \<subseteq> old: Orderings.ordering
  by (unfold_locales, auto simp: strict_as_weak dest: trans)

locale linear_ordering = total + ordering
begin

sublocale total_quasi_order + total_pseudo_order..

sublocale strict: linear_irreflexive by (unfold_locales, auto simp: strict_as_weak not_weak_iff)

lemma not_weak[simp]: "\<not>x \<sqsubseteq> y \<longleftrightarrow> y \<sqsubset> x" by (simp add: not_weak_iff strict_as_weak)

lemma not_strict[simp]: "\<not>x \<sqsubset> y \<longleftrightarrow> y \<sqsubseteq> x" by (auto simp: strict_as_weak)

end


text {* A locale which defines an equivalence relation. Be careful when declaring simp rules etc.,
  as the equivalence will often be rewritten to equality. *}

locale quasi_order_equivalence = quasi_order + equivalence_syntax +
  assumes equiv_def: "x \<sim> y \<longleftrightarrow> x \<sqsubseteq> y \<and> y \<sqsubseteq> x"
begin

sublocale equiv: equivalence by (unfold_locales, auto simp: equiv_def dest: trans)

lemma [trans]:
  shows equiv_weak_trans: "x \<sim> y \<Longrightarrow> y \<sqsubseteq> z \<Longrightarrow> x \<sqsubseteq> z"
    and weak_equiv_trans: "x \<sqsubseteq> y \<Longrightarrow> y \<sim> z \<Longrightarrow> x \<sqsubseteq> z"
  by (auto simp: equiv_def dest: trans)

lemma extreme_order_iff_equiv[simp]: "extreme (\<sqsubseteq>) {x. x \<sqsubseteq> e} y \<longleftrightarrow> e \<sim> y"
  by (auto simp: equiv_def intro!: extremeI dest: trans)

lemma extreme_bound_iff_equiv:
  assumes bX: "extreme_bound (\<sqsubseteq>) X b" shows "extreme_bound (\<sqsubseteq>) X c \<longleftrightarrow> b \<sim> c"
proof(rule iffI)
  from bX have bX: "bound (\<sqsubseteq>) X b" and leastb: "\<And>x. bound (\<sqsubseteq>) X x \<Longrightarrow> b \<sqsubseteq> x" by auto
  { fix c assume "extreme_bound (\<sqsubseteq>) X c"
    then have cbounds: "bound (\<sqsubseteq>) X c" and leastc: "\<And>b. bound (\<sqsubseteq>) X b \<Longrightarrow> c \<sqsubseteq> b" by auto
    from leastb[OF cbounds] leastc[OF bX] show "b \<sim> c" by (auto simp: equiv_def)
  }
  { fix c assume bc: "b \<sim> c"
    show "extreme_bound (\<sqsubseteq>) X c"
    proof(intro extreme_boundI)
      fix x assume "x \<in> X"
      with bX have "x \<sqsubseteq> b" by auto
      with bc show "x \<sqsubseteq> c" by (auto dest: trans simp: equiv_def)
    next
      fix x assume "bound (\<sqsubseteq>) X x"
      from leastb[OF this] bc show "c \<sqsubseteq> x" by (auto dest: trans simp: equiv_def)
    qed
  }
qed

lemma extremes_are_equiv: "extreme (\<sqsubseteq>) X x \<Longrightarrow> extreme (\<sqsubseteq>) X y \<Longrightarrow> x \<sim> y"
  by (auto simp: equiv_def)

end

locale quasi_ordering_equivalence = compatible_ordering + quasi_order_equivalence
begin

lemma [trans]:
  shows equiv_strict_trans: "x \<sim> y \<Longrightarrow> y \<sqsubset> z \<Longrightarrow> x \<sqsubset> z"
    and strict_equiv_trans: "x \<sqsubset> y \<Longrightarrow> y \<sim> z \<Longrightarrow> x \<sqsubset> z"
  by (auto simp: equiv_def dest: weak_strict_trans strict_weak_trans)

end

subsection {* Relating to Classes *}

text {* In Isabelle/2018, we should declare sublocales in class before declaring dual version,
  since otherwise we facts are prefixed by dual.dual.
*}

context ord begin

abbreviation upper_bound where "upper_bound \<equiv> bound (\<le>)"

abbreviation least where "least \<equiv> extreme (\<ge>)"

abbreviation lower_bound where "lower_bound \<equiv> bound (\<lambda>x y. y \<le> x)"

abbreviation greatest where "greatest \<equiv> extreme (\<lambda>x y. y \<ge> x)"

end

lemma fun_ord_le[simp]: "fun_ord (\<le>) = (\<le>)" by (intro ext, simp add: fun_ord_def le_fun_def)
lemma fun_ord_ge[simp]: "fun_ord (\<ge>) = (\<ge>)" by (intro ext, simp add: fun_ord_def le_fun_def)

lemmas fun_lub_iff = fun_extreme_bound_iff[of "(\<le>)", unfolded fun_ord_le]
lemmas fun_glb_iff = fun_extreme_bound_iff[of "(\<ge>)", unfolded fun_ord_ge]


class compat = ord + assumes "compatible_ordering (\<le>) (<)"
begin

sublocale ord: compatible_ordering using compat_axioms unfolding class.compat_def.

end

class psorder = ord + assumes "pseudo_ordering (\<le>) (<)"
begin

text {* We need to declare subclasses before sublocales. *}

interpretation ord: pseudo_ordering using psorder_axioms unfolding class.psorder_def.

subclass compat..

sublocale ord: pseudo_ordering..

lemma Least_eq_The_least: "Least P = The (least {x. P x})"
  unfolding Least_def
  unfolding extreme_def[unfolded atomize_eq, THEN ext]
  by auto

text {* We postpone @{term Greatest} since, strangely, it is available only on class order. *}

end

class qorder = ord + assumes "quasi_ordering (\<le>) (<)"
begin

interpretation ord: quasi_ordering using qorder_axioms unfolding class.qorder_def.

subclass compat..

sublocale ord: quasi_ordering..

end

text {* Isabelle/HOL's @{class preorder} belongs to @{class qorder}. *}
subclass (in preorder) qorder
  apply unfold_locales
  apply (fact order_refl)
  apply simp
  apply (fact le_less_trans)
  apply (fact less_le_trans)
  apply (fact less_imp_le)
  apply (fact order_trans)
  done

class worder = ord + assumes "partial_ordering (\<le>) (<)"
begin

interpretation ord: partial_ordering using worder_axioms unfolding class.worder_def.

subclass psorder..
subclass qorder..

sublocale ord: partial_ordering..

lemma Greatest_eq_The_greatest: "Greatest P = The (greatest {x. P x})"
  unfolding Greatest_def
  unfolding extreme_def[unfolded atomize_eq, THEN ext]
  by auto

end

text {* Isabelle/HOL's @{class order} belongs to @{class worder}. *}
subclass (in order) worder by (unfold_locales, auto)

class wf_qorder = ord + assumes "well_founded_ordering (\<le>) (<)"
begin

interpretation ord: well_founded_ordering using wf_qorder_axioms unfolding class.wf_qorder_def.

subclass qorder..

sublocale ord: well_founded_ordering..

end

subclass (in wellorder) wf_qorder by (unfold_locales, fact less_induct)

context linorder begin

sublocale ord: linear_ordering by (unfold_locales, auto)

end

subsection {* Declaring Duals *}

sublocale less_eq_syntax \<subseteq> dual: less_eq_syntax "(\<sqsubseteq>)\<inverse>\<inverse>".

sublocale reflexive \<subseteq> dual: reflexive "(\<sqsubseteq>)\<inverse>\<inverse>" by auto

sublocale irreflexive \<subseteq> dual: irreflexive "(\<sqsubset>)\<inverse>\<inverse>" by (unfold_locales, auto)

sublocale transitive \<subseteq> dual: transitive "(\<sqsubseteq>)\<inverse>\<inverse>" by (unfold_locales, erule trans)

sublocale antisymmetric \<subseteq> dual: antisymmetric "(\<sqsubseteq>)\<inverse>\<inverse>" by auto

sublocale asymmetric \<subseteq> dual: asymmetric "(\<sqsubset>)\<inverse>\<inverse>" by unfold_locales

sublocale total \<subseteq> dual: total "(\<sqsubseteq>)\<inverse>\<inverse>" using total by auto

sublocale linear \<subseteq> dual: linear "(\<sqsubset>)\<inverse>\<inverse>" using linear by auto

sublocale pseudo_order \<subseteq> dual: pseudo_order "(\<sqsubseteq>)\<inverse>\<inverse>" ..

sublocale quasi_order \<subseteq> dual: quasi_order "(\<sqsubseteq>)\<inverse>\<inverse>" ..

sublocale partial_order \<subseteq> dual: partial_order "(\<sqsubseteq>)\<inverse>\<inverse>"..

text {*
  In the following dual sublocale declaration, ``rewrites'' eventually cleans up redundant facts.
*}
sublocale symmetric \<subseteq> dual: symmetric "(\<sim>)\<inverse>\<inverse>" rewrites "(\<sim>)\<inverse>\<inverse> = (\<sim>)"
  using symmetric_axioms by (auto simp: dual_sym)

sublocale equivalence \<subseteq> dual: equivalence "(\<sim>)\<inverse>\<inverse>" rewrites "(\<sim>)\<inverse>\<inverse> = (\<sim>)"
  by (unfold_locales, auto simp: dual_sym sym)

sublocale linear_irreflexive \<subseteq> dual: linear_irreflexive "(\<sqsubset>)\<inverse>\<inverse>"..

sublocale total_pseudo_order \<subseteq> dual: total_pseudo_order "(\<sqsubseteq>)\<inverse>\<inverse>"..

sublocale total_quasi_order \<subseteq> dual: total_quasi_order "(\<sqsubseteq>)\<inverse>\<inverse>" ..

sublocale compatible_ordering \<subseteq> dual: compatible_ordering "(\<sqsubseteq>)\<inverse>\<inverse>" "(\<sqsubset>)\<inverse>\<inverse>"
  using weak_strict_trans strict_weak_trans strict_implies_weak by unfold_locales

sublocale pseudo_ordering \<subseteq> dual: pseudo_ordering "(\<sqsubseteq>)\<inverse>\<inverse>" "(\<sqsubset>)\<inverse>\<inverse>"..

sublocale quasi_ordering \<subseteq> dual: quasi_ordering "(\<sqsubseteq>)\<inverse>\<inverse>" "(\<sqsubset>)\<inverse>\<inverse>"..

sublocale partial_ordering \<subseteq> dual: partial_ordering "(\<sqsubseteq>)\<inverse>\<inverse>" "(\<sqsubset>)\<inverse>\<inverse>"..

sublocale ordering \<subseteq> dual: ordering "(\<sqsubseteq>)\<inverse>\<inverse>" "(\<sqsubset>)\<inverse>\<inverse>" by (unfold_locales, auto simp: strict_as_weak)

sublocale linear_ordering \<subseteq> dual: linear_ordering "(\<sqsubseteq>)\<inverse>\<inverse>" "(\<sqsubset>)\<inverse>\<inverse>" ..


lemma(in antisymmetric) monotone_extreme_imp_extreme_bound_iff:
  fixes ir (infix "\<preceq>" 50)
  assumes "monotone (\<preceq>) (\<sqsubseteq>) f" and i: "extreme (\<preceq>) C i"
  shows "extreme_bound (\<sqsubseteq>) (f ` C) x \<longleftrightarrow> f i = x"
  using dual.extreme_unique monotone_extreme_extreme_boundI[OF assms] by auto


subsection {* Instantiations *}

text {*
Pointwise order pair of functions are compatible only if the weak part is transitive.
*}

instance "fun" :: (type,qorder) qorder
proof (intro_classes, unfold_locales)
  note [simp] = le_fun_def less_fun_def
  fix f g h :: "'a \<Rightarrow> 'b"
  { assume fg: "f \<le> g" and gh: "g < h"
    show "f < h"
    proof (unfold less_fun_def, intro conjI le_funI notI)
      from fg have "f x \<le> g x" for x by auto
      also from gh have "g x \<le> h x" for x by auto
      finally show "f x \<le> h x" for x.
      assume hf: "h \<le> f"
      then have "h x \<le> f x" for x by auto
      also from fg have "f x \<le> g x" for x by auto
      finally have "h \<le> g" by auto
      with gh show False by auto
    qed
  }
  { assume fg: "f < g" and gh: "g \<le> h"
    show "f < h"
    proof (unfold less_fun_def, intro conjI le_funI notI)
      from fg have "f x \<le> g x" for x by auto
      also from gh have "g x \<le> h x" for x by auto
      finally show "f x \<le> h x" for x.
      assume hf: "h \<le> f"
      then have "h x \<le> f x" for x by auto
      also from gh have "g x \<le> h x" for x by auto
      finally have "g \<le> f" by auto
      with fg show False by auto
    qed
  }
  show "f < g \<Longrightarrow> f \<le> g" by auto
  show "\<not>f < f" by auto
qed (auto intro!: le_funI elim: le_funE dest: ord.trans)

instance "fun" :: (type,worder) worder
  by (intro_classes, unfold_locales, auto simp: less_fun_def le_fun_def)

instance nat :: linorder..

end
